@@ -19,10 +19,13 @@
 #define TILEDB_DEPRECATED_EXPORT
 
 #include <tiledb/arrowio>
-#include <tiledb/tiledb.h>               // C
-#include <tiledb/tiledb>                 // C++
-#include <tiledb/tiledb_experimental.h>  // C
+#include <tiledb/tiledb.h>              // C
+#include <tiledb/tiledb>                // C++
+#include <tiledb/tiledb_experimental.h> // C
+
+#if defined(TILEDB_SERIALIZATION)
 #include <tiledb/tiledb_serialization.h> // C
+#endif
 
 #include "../external/string_view.hpp"
 #include "../external/tsl/robin_map.h"
@@ -287,6 +290,7 @@ private:
   bool exact_init_bytes_ = false;
   uint64_t init_buffer_bytes_ = DEFAULT_INIT_BUFFER_BYTES;
   uint64_t alloc_max_bytes_ = DEFAULT_ALLOC_MAX_BYTES;
+  tiledb_layout_t layout_ = TILEDB_ROW_MAJOR;
 
   py::object pyschema_;
 
@@ -361,11 +365,11 @@ public:
         std::shared_ptr<tiledb::Query>(new Query(ctx_, *array_, TILEDB_READ));
     //        [](Query* p){} /* note: no deleter*/);
 
-    tiledb_layout_t layout = (tiledb_layout_t)py_layout.cast<int32_t>();
-    if (!issparse && layout == TILEDB_UNORDERED) {
+    layout_ = (tiledb_layout_t)py_layout.cast<int32_t>();
+    if (!issparse && layout_ == TILEDB_UNORDERED) {
       TPY_ERROR_LOC("TILEDB_UNORDERED read is not supported for dense arrays")
     }
-    query_->set_layout(layout);
+    query_->set_layout(layout_);
 
 #if TILEDB_VERSION_MAJOR >= 2 && TILEDB_VERSION_MINOR >= 2
     if (use_arrow_) {
@@ -600,7 +604,9 @@ public:
     for (auto dim_range : ranges) {
       // py::print(dim_range);
       if (py::isinstance<py::array>(dim_range)) {
-        py::array r_array = dim_range.cast<py::array>();
+        auto tiledb_type = domain_->dimension(dim_idx).type();
+        py::dtype dtype = tiledb_dtype(tiledb_type, 1);
+        py::array r_array = dim_range.attr("astype")(dtype);
         add_bulk_range(dim_idx, r_array);
       } else {
         py::tuple dim_range_iter = dim_range.cast<py::iterable>();
@@ -652,6 +658,7 @@ public:
     }
   }
 
+#if defined(TILEDB_SERIALIZATION)
   void set_serialized_query(py::buffer serialized_query) {
     int rc;
     tiledb_query_t *c_query;
@@ -673,6 +680,7 @@ public:
     if (rc == TILEDB_ERR)
       TPY_ERROR_LOC("Could not deserialize query.");
   }
+#endif
 
   void set_attr_cond(py::object attr_cond) {
     if (!attr_cond.is(py::none())) {
@@ -817,6 +825,11 @@ public:
       } else { // !nullable && !var
         buf_nbytes = query_->est_result_size(name);
       }
+
+      // Add extra offset to estimate in order to avoid incomplete resubmit
+      // libtiledb 2.7.* does not include extra element in estimate.
+      // Remove this section after resolution of SC-16301.
+      offsets_num += (var && use_arrow_) ? 1 : 0;
     }
 
     // - for sparse arrays: don't try to allocate more than alloc_max_bytes_
@@ -849,13 +862,16 @@ public:
   }
 
   py::object get_buffers() {
-    py::list result;
+    py::dict rmap;
     for (auto &bp : buffers_) {
+      py::list result;
+      const auto name = bp.first;
       const BufferInfo b = bp.second;
       result.append(b.data);
       result.append(b.offsets);
+      rmap[py::str{name}] = result;
     }
-    return std::move(result);
+    return std::move(rmap);
   }
 
   void set_buffers() {
@@ -1397,11 +1413,35 @@ public:
   }
 
   py::object estimated_result_sizes() {
-    py::dict results;
-    for (auto const &bp : buffers_) {
-      auto name = bp.first;
-      auto buf = bp.second;
+    // vector of names to estimate
+    std::vector<std::string> estim_names;
 
+    for (size_t dim_idx = 0; dim_idx < domain_->ndim(); dim_idx++) {
+      auto dim = domain_->dimension(dim_idx);
+      if ((std::find(dims_.begin(), dims_.end(), dim.name()) == dims_.end()) &&
+          // we need to also check if this is an attr for backward-compatibility
+          (std::find(attrs_.begin(), attrs_.end(), dim.name()) ==
+           attrs_.end())) {
+        continue;
+      }
+      estim_names.push_back(dim.name());
+    }
+
+    // iterate by idx: schema.attributes() is unordered, but we need to
+    //                 return ordered results
+    for (size_t attr_idx = 0; attr_idx < array_schema_->attribute_num();
+         attr_idx++) {
+      auto attr = array_schema_->attribute(attr_idx);
+      if (std::find(attrs_.begin(), attrs_.end(), attr.name()) ==
+          attrs_.end()) {
+        continue;
+      }
+      estim_names.push_back(attr.name());
+    }
+
+    py::dict results;
+
+    for (auto const &name : estim_names) {
       size_t est_offsets = 0, est_data_bytes = 0;
 
       if (is_var(name)) {
@@ -1521,7 +1561,9 @@ void init_core(py::module &m) {
 #endif
           .def("set_subarray", &PyQuery::set_subarray)
           .def("set_attr_cond", &PyQuery::set_attr_cond)
+#if defined(TILEDB_SERIALIZATION)
           .def("set_serialized_query", &PyQuery::set_serialized_query)
+#endif
           .def("submit", &PyQuery::submit)
           .def("unpack_buffer", &PyQuery::unpack_buffer)
           .def("estimated_result_sizes", &PyQuery::estimated_result_sizes)
